#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May  7 14:58:53 2019

@author: mpolikarpov
"""


import numpy as np
import matplotlib.pyplot as plt
import scipy
import cv2



class polar_image:
    '''
    description
    '''
    
# =============================================================================
#     def __init__(self, data, origin = None):
#         """Constructor"""
#         self.data = data
#         self.origin = origin
# =============================================================================
    def __init__(self):
        """Constructor"""
        pass

    # =============================================================================
    # some functions for polar interpolation
    # =============================================================================
    
    def plot_polar_image(self, data, origin=None):
        """Plots an image reprojected into polar coordinages with the origin
        at "origin" (a tuple of (x0, y0), defaults to the center of the image)"""
        polar_grid, r, theta = self.reproject_image_into_polar(data, origin)
        plt.figure()
        plt.imshow(polar_grid, extent=(theta.min(), theta.max(), r.max(), r.min()))
        plt.axis('auto')
        plt.ylim(plt.ylim()[::-1])
        plt.xlabel('Theta Coordinate (radians)')
        plt.ylabel('R Coordinate (pixels)')
        plt.title('Image in Polar Coordinates')


    def index_coords(self, data, origin=None):
        """Creates x & y coords for the indicies in a numpy array "data".
        "origin" defaults to the center of the image. Specify origin=(0,0)
        to set the origin to the lower left corner of the image."""
        ny, nx = data.shape[:2]
        if origin is None:
            origin_x, origin_y = nx // 2, ny // 2
        else:
            origin_x, origin_y = origin
        x, y = np.meshgrid(np.arange(nx), np.arange(ny))
        x -= origin_x
        y -= origin_y
        return x, y
    
    
    def cart2polar(self, x, y):
        r = np.sqrt(x**2 + y**2)
        theta = np.arctan2(y, x)
        return r, theta
    
    
    def polar2cart(self, r, theta):
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y
    
    
    def reproject_image_into_polar(self, data, origin=None):
        """Reprojects a 3D numpy array ("data") into a polar coordinate system.
        "origin" is a tuple of (x0, y0) and defaults to the center of the image."""
        # horizontal direction on the image will be theta coordinate 
        #(with 0 in the image center), vertical - R coordinate (with 0 in left top corner)
        
        data = cv2.merge([data,data,data])
        ny, nx = data.shape[:2]
        if origin is None:
            origin = (nx//2, ny//2)
    
        # Determine that the min and max r and theta coords will be...
        x, y = self.index_coords(data, origin=origin)
        r, theta = self.cart2polar(x, y)
    
        # Make a regular (in polar space) grid based on the min and max r & theta
        r_max = int(round(np.sqrt((nx//2)**2 + (ny//2)**2)))
        r_i = np.linspace(r.min(), r.max(), r_max)
        theta_i = np.linspace(theta.min(), theta.max(), 360)
        theta_grid, r_grid = np.meshgrid(theta_i, r_i)
    
        # Project the r and theta grid back into pixel coordinates
        xi, yi = self.polar2cart(r_grid, theta_grid)
        xi += origin[0] # We need to shift the origin back to 
        yi += origin[1] # back to the lower-left corner...
        xi, yi = xi.flatten(), yi.flatten()
        coords = np.vstack((xi, yi)) # (map_coordinates requires a 2xn array)
    
        # Reproject each band individually and the restack
        # (uses less memory than reprojection the 3-dimensional array in one step)
        bands = []
        for band in data.T:
            zi = scipy.ndimage.map_coordinates(band, coords, order=1)
            bands.append(zi.reshape((r_max, 360)))
        output = np.dstack(bands)
        return output#, r_i, theta_i
    
    
    def baseline(self, data1D, polynom = 5):
        '''
        substract the baseline from 2D curve
        '''
        x = np.linspace(0, data1D.shape[0], len(data1D), dtype = 'uint16')
        y = data1D
        
        pfit = np.polyfit(x, y, polynom)                         # returns polynomial coefficients
        yp = np.polyval(pfit, x)                                # fits the curve,
        smoothed = data1D - yp                                  # subtract baseline
    
        return smoothed


    


"""
Created on Fri Oct  12 15:46:49 2018

@author: mpolikarpov
"""


from numpy import floor, meshgrid, multiply, divide, square, sin, cos, sqrt, mean, pi, fliplr, flipud, rot90, ceil, arange, concatenate
import numpy
from numpy.fft import ifftshift, fft2, ifft2
from scipy.special import erfc
from skimage.feature import register_translation
from scipy.ndimage import fourier_shift
from maximus48.monochromaticCTF import single_distance_CTF as sdCTF








def shift_distance(image1, image2, accuracy = 100):
    """
    Finds lateral shift between two images 
    
    Parameters
    __________
    image1 : 2D array

    image2 : 2D array
        y axis
    accuracy: int
        Upsampling factor. Images will be registered within 1 / upsample_factor of a pixel. For example upsample_factor == 20 means the images will be registered within 1/20th of a pixel.    
    """
    
    shift, error, diffphase = register_translation(image1, image2, 100)
    
    return shift



def find_shift_CTF(image1, image2, beta_delta, fresnelNumber1, fresnelNumber2, accuracy=100, zero_compensation=0.01):
    """
    Finds the lateral shift between two holographic projections by reconstructing the phase
    
    Parameters
    __________
    image1 : 2D array

    image2 : 2D array
        y axis
    accuracy: int
        Upsampling factor. Images will be registered to within 1 / upsample_factor of a pixel. For example upsample_factor == 20 means the images will be registered within 1/20th of a pixel.    
    beta_delta: int
        absorption divided by refraction  (decrements). Normally small, for ex. 0.08 
    fresnel Number(1,2): float 
        fresnel numbers for corresponding pictures
        can be calculated as fresnelN = pixelsize**2/(wavelength*object_detector_distance)
    zero_compensation: float
        just a parameter, should be small, for ex. 0.01
    """
        
    image1_CTF = sdCTF(image1, beta_delta, fresnelNumber1, zero_compensation) 
    image2_CTF = sdCTF(image2, beta_delta, fresnelNumber2, zero_compensation) 
        
    shift = shift_distance(image1_CTF, image2_CTF, accuracy)
    
    return shift



def shift_image(image, shift):
    """
    Laterally shifts one image. 
    
    Parameters
    __________
    image : 2D array
    shift: float
        shift value
    """
    
    back_image = fourier_shift(numpy.fft.fftn(image), shift)
    back_image = numpy.fft.ifftn(back_image).real
    return back_image



def shift_imageset(image, shift):
    """
    Aligns every image in the image-set by lateral shift. 
    
    Parameters
    __________
    image : 3D array
        first axis is images acquired at different distances
    shift: 3D aray 
        shift values for every image
    """
    #out = numpy.zeros(image.shape, dtype = image.dtype)
    for i in numpy.arange(len(image)):
        image[i] = shift_image(image[i], shift[i])
    
    return image



# =============================================================================
# def shift_imageset(image, shift):
#     """
#     Aligns every image in the image-set by lateral shift. 
#     
#     Parameters
#     __________
#     image : 3D array
#         first axis is images acquired at different distances
#     shift: 3D aray 
#         shift values for every image
#     """
#     out = numpy.zeros(image.shape, dtype = image.dtype)
#     for i in numpy.arange(len(image)):
#         out[i] = shift_image(image[i], shift[i])
#     
#     return out
# 
# =============================================================================


# =============================================================================
# 
# 
# def mdCTF_imageset(image, beta_delta, fresnelN, accuracy, zero_compensation, Npad = 200):
#     """
#     Shifts images and returns result of multi-distance CTF-retrieval
#     
#     Parameters
#     __________
#     image : 3D array
#         first axis is images acquired at different distances
#     beta_delta: int
#         absorption divided by refraction  (decrements). Normally small, for ex. 0.08 
#     fresnel Number(1,2): array 
#         fresnel numbers for corresponding pictures
#         can be calculated as fresnelN = pixelsize**2/(wavelength*object_detector_distance)
#     accuracy: int
#         Upsampling factor. Images will be registered to within 1 / upsample_factor of a pixel. For example upsample_factor == 20 means the images will be registered within 1/20th of a pixel.    
#     beta_delta: int
#         absorption divided by refraction  (decrements). Normally small, for ex. 0.08 
#     fresnelN: tuple 
#         fresnel numbers for corresponding pictures, can be calculated as fresnelN = pixelsize**2/(wavelength*object_detector_distance)
#     zero_compensation: float
#         just a parameter, should be small, for ex. 0.01
#     Npad : int
#         edge padding before phase-retrieval
#     """
#     
#     out = numpy.zeros(image.shape, dtype = image.dtype)
#     
#     #shift images
#     for i in numpy.arange(out.shape[0]):
#         out[i] = shift_image_CTF(image[0], image[i], beta_delta, fresnelN[0], fresnelN[i], accuracy, zero_compensation)
#     
#     
#     
#     
#     
#     #pad images
#     out = numpy.pad(out, ((0,0),(Npad, Npad),(Npad, Npad)), 'edge')        
#     # multiple-distance CTF-retrieval
#     out = multi_distance_CTF(out, beta_delta, fresnelN, zero_compensation)
#     #unpad images
#     out = out[Npad:(out.shape[0]-Npad),Npad:(out.shape[1]-Npad)]
#     
#     return out
# =============================================================================
















def multi_distance_CTF(image, beta_delta, fresnelNumber, zero_compensation):
    """
    This function makes the CTF phase-retrieval from the multiple-distance images.
    Please refer to the book of Tim Salditt "Biomedical imaging", 2017, ISBN = 978-3-11-042668-7, ISBN (PDF)= 978-3-11-0426694
    equation 6.144
    
    Parameters
    __________
    image : 3d array, float32
        0 axis is for different images that you want to CTF_calculate
    beta_delta: int
        absorption divided by refraction  (decrements). Normally small, for ex. 0.08 
    fresnel Number: array 
        fresnel numbers for corresponding pictures
        can be calculated as fresnelN = pixelsize**2/(wavelength*object_detector_distance)
    zero_compensation: int
        just a parameter, should be small, for ex. 0.01
    """
    
    nom = numpy.zeros((image.shape[1],image.shape[2]), dtype = 'complex128')
    denom = numpy.zeros((image.shape[1],image.shape[2]), dtype = 'complex128')
    
    for i in arange(image.shape[0]):
        nominator, denominator, f = ctf_mono((image.shape[1],image.shape[2]), 
                                             fresnelNumber[i], fresnelNumber[i], 
                                             beta_delta, zero_compensation)
        nom += multiply(fft2(image[i,:,:]-1), nominator) 
        denom += denominator
        
    ctf_out = multiply(divide(nom, denom), f)  
    ctf_out[abs(ctf_out)<zero_compensation] = zero_compensation 

    reco = ifft2(ctf_out).real
   
    return reco
    


def circle(x,y,r):
    """
    This function creates a mask - 2D array with zeros outside a circular region and ones inside the circle.
    x is the xaxis of the coordinate system, y is the y axis. The third argument is the radius of the circle. 
    
    Parameters
    __________
    x : 1D array
        x axis
    y : 1D array
        y axis
    r: int
        radius of the mask circle    
    """
    
    X,Y = meshgrid(x,y)
    c = sqrt(square(X)+square(Y))
    c[c<=r]=1
    c[c>r]=0
    return c 



def ctf_mono(imageshape, fresnelx, fresnely, beta_delta, zero_compensation):
    """
    This is the main function which calculates nominator/denominator. Please refer to the book of Tim Salditt - Biomedical imaging (2017)
    """
    # set fourier coordinate system
    Ny = imageshape[0]
    Nx = imageshape[1] 
    dqx = 1/Nx
    dqy = 1/Ny
    
    xx = arange(1,Nx+1)-floor(Nx/2)-1 
    yy = arange(1,Ny+1)-floor(Ny/2)-1
    Qx, Qy = meshgrid(xx*dqx, yy*dqy)
      
    # formulate the contrast transfer function
    argument = ifftshift(pi/fresnelx*square(Qx)+pi/fresnely*square(Qy))         # Салдит в своей книге по другому задает Qx, в 2пи раз больше
    ctf_mono_nominator = sin(argument)+ beta_delta * cos(argument)
    ctf_mono_denominator =  2 * square(ctf_mono_nominator)  
    
    # regularization    
    # cutoff matrix using the function from P. Cloetens
    cutoff_sigma = 0.01
    cutoff = sqrt(mean(fresnelx)*0.5)
    r = filt2(Ny, cutoff*Ny, cutoff_sigma*Ny, Nx, cutoff*Nx)  
    regularization = multiply(zero_compensation,(1-r))              
    ctf_mono_denominator = ctf_mono_denominator + regularization  
    
    # aliasing free region f
    N_correct = min(floor(fresnelx / dqx**2), floor(fresnely / dqy**2)) 
    f  =  ifftshift(circle(xx,yy,floor(N_correct/2)))
     
    return ctf_mono_nominator, ctf_mono_denominator, f
    


def filt2(n, cutn, sigman, m=None, cutm=None):
        
# =============================================================================
# # regularisation function
# =============================================================================

# =============================================================================
# % Copyright (C) 2007 P. Cloetens
# %
# % This program is free software; you can redistribute it and/or modify
# % it under the terms of the GNU General Public License as published by
# % the Free Software Foundation; either version 2 of the License, or
# % (at your option) any later version.
# %
# % This program is distributed in the hope that it will be useful,
# % but WITHOUT ANY WARRANTY; without even the implied warranty of
# % MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# % GNU General Public License for more details.
# %
# % You should have received a copy of the GNU General Public License
# % along with this program; if not, write to the Free Software
# % Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
# %
# % filt2
# %		r = filt2(n,cutn,sigman,varargin)
# %		create low-pass 2D filter in Fourier domain
# %		erfc type
# %
# %		arguments:
# %		argument 1: number of pixels along dimension 1
# %		argument 2: cut_off along dimension 1
# %		argument 3: width (sigma) of transition along dimension 1
# %		argument 4: number of pixels along dimension 2 ( default : same as argument 1)
# %		argument 5: cut_off along dimension 2 ( default : isotropic )
# %
# %		See also: filt2_gaussian
# 
# % Author: P. Cloetens <cloetens@esrf.fr>
# %
# % 2007-03-26 P. Cloetens <cloetens@esrf.fr>
# % * Initial revision
# %
# % adapted to Matlab: Martin Krenkel, February 2013
# =============================================================================
    
    if not m:
        m = n
    if not cutm:    
        cutm = cutn*m/n
        
        
    x,y = meshgrid(arange(floor(m/2)+1),arange(floor(n/2)+1))
    r = sqrt(square(x)+square(y*(cutm/cutn)))
 
    # im Prinzip nur hier:
    r = erfc((r-cutn)/sigman)
    r=r/r[0,0]
    
    r1 = fliplr(r[:,1:int(ceil(m/2))])
    r2 = flipud(r[1:int(ceil(n/2)),:]) 
    r3 = rot90(r[1:int(ceil(n/2)),1:int(ceil(m/2))],2)
    
    out0 = concatenate((r,r1), axis=1)
    out1 = concatenate((r2,r3), axis=1)
    out  = concatenate((out0,out1), axis=0)
        
    return out










#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May  7 14:58:53 2019

@author: mpolikarpov

Замечанания от 171019
На примере образца номер 5_3 из данных Бенедикта (Бремен) с эксперимента на P14_EMBL от 220919
если не восст фазу то центр вращения 872.475 и 872.135 (старый алгоритм - когда просто сравниваются две картинки как целое)
если есть восст фазы то центр 1384.5525 и 1384.505 (старый алгоритм) - в пересчете это 872.5525 и 872.5
выввод - обрезание картинки вообще говоря уменьшает разброс между старым и новым алогоритмом (наверное логично так как уменьшается шум)

На примере образца 5_2 
я увидел что восттановление фазы вносит бОльшую погрешность в определение центра вращения
так что лучше находить ось вращения на RAW изображениях но с флетфилд-коррекцией
если не восст фазу то центр вращения 870.017 и 869.83 (старый алгоритм)



Общий вывод - находи ось на картинках с flatfield-correction но без восстановления фазы
Края картинок можешь чуть обрезать
"""


import numpy as np
import matplotlib.pyplot as plt
import os
from maximus48.multiCTF2 import shift_distance as shift
from multiprocessing import Pool 








class rotation:
    '''
    description
    '''
    
# =============================================================================
#     def __init__(self, data, origin = None):
#         """Constructor"""
#         self.data = data
#         self.origin = origin
# =============================================================================
    def __init__(self):
        """Constructor"""
        pass
    
    
    
    

    # =============================================================================
    # some functions to effectively fint the rotation axis in the tomography experiment
    # =============================================================================
    
    
    def axis_raws(self, image1, image2, Npad = 0, RotROI = None, level = 5, window = 50):
        """Finds an axis of rotation by comparing the 1st and the 180deg image:
        image1: 2D array 
            first 2D image
        image2: 2D array
            second 2D image
        Npad: int 
            type the number here if the x axis of your image was padded
        RotROI: tuple
            ROI where to compare the images. Note that it is better to exclude some regions at the edge of the camera
        level: int 
            the number of pixels to be taken into account, please type None if you don't want to use it and want to export the whole data.
        window: int
            window is the number of pixels on the image (height) to be taken into account during comaprison
        """
        
        if not RotROI:
            RotROI = (0, image1.shape[0],
                      Npad + image1.shape[1]//8, image1.shape[1] - Npad - image1.shape[1]//8)
        
    
        all_cent=[]
        for i in range(RotROI[0], RotROI[1], window//2):
    
            im_1 = image1[i:i+window,RotROI[2]:RotROI[3]]
            im_2 = np.flip(image2[i:i+window,RotROI[2]:RotROI[3]],1)
    
            distances = shift(im_1, im_2)
            cent = im_1.shape[1]/2 + distances[1]/2 + RotROI[2]
    
            all_cent.append(cent)
            #print('center for the slice #',i,' is: ', cent)


                
        if level:
            x=[]
            y=[]

            for i in range(len(all_cent)):
                if np.absolute(all_cent[i] - np.median(all_cent)) <level:
                    x.append(i * window//2)
                    y.append(all_cent[i])
        
            all_cent = np.column_stack((x,y))
        
        else:
            all_cent = np.column_stack((np.linspace(0, len(all_cent), len(all_cent), dtype = 'uint16'), all_cent))
        
        #plt.plot(all_cent[:,0], all_cent[:,1], 'bo')
                    
        return all_cent
    
    




    def interpolate(self, cent, level = None):
        """interpolates the coordinates for the rotation axis with the line
         basically finds the inclination of the rotation axis through the image
       
        cent: 2D array 
            comes from axis_raws    
        level:int 
            is the number of pixels to be taken into account if you want 
            to discard all values that have more than 5 degrees difference with the median
        """  
    
        step = cent[1,0] - cent[0,0]
        
        if not level:
            x = cent[:,0]
            y = cent[:,1]
        
        else:
            x = []
            y = []
            for i in range(len(cent)):
                if np.absolute(cent[i,1] - np.median(cent[:,1])) < level:
                    x.append(i*step)
                    y.append(cent[i,1])
                    
        
        pfit = np.polyfit(x, y, 1)                                                       # returns polynomial coefficients
        #yp = np.polyval(pfit, x)                                                         # fits the curve,
        
        return pfit
        
        



        
    
        
            
# =============================================================================
#     def axis_raws_parallel(self, image1, image2, Npad = 200, RotROI = None, level = 5, ncore = 10):
#         """Finds an axis of rotation by comparing the 1st and the 180deg image:
#         image1 - first 2D image
#         image2 - second 2D image
#         Npad - type the number here if the x axis of your image was padded
#         RotROI - ROI where to compare the images. Note that it is better to exclude some regions at the edge of the camera
#         level - is the number of pixels to be taken into account, please type None if you don't want to use it and want to export the whole data.
#             """
#         
#         if not RotROI:
#             RotROI = (0, image1.shape[0],
#                       Npad+image1.shape[1]//8, image1.shape[1]-Npad-image1.shape[1]//8)
#         
#     
#         all_cent=[]
# 
#         im_1 = image1[:,RotROI[2]:RotROI[3]]
#         im_2 = np.flip(image2[:,RotROI[2]:RotROI[3]],1)
#         
# # =============================================================================
# #         def process(i):
# #     
# #             distances = shift(im_1[i:i+1], im_2[i:i+1])
# #             cent = im_1.shape[1]/2 + distances[1]/2 + RotROI[2]
# #     
# #             return cent
# # =============================================================================
# 
# 
#         os.environ['OMP_NUM_THREADS'] ='1'
#         with Pool(ncore) as pool:    
#             all_cent = pool.map(process, np.arange(RotROI[0], RotROI[1]))
#         os.environ['OMP_NUM_THREADS'] ='100'
# 
# 
# 
#     def center(i):
#         distances = shift(im_1[i:i+1], im_2[i:i+1])
#         cent = im_1.shape[1]/2 + distances[1]/2
#         
#         return cent
#         
#         
#                 
# # =============================================================================
# #         if level:
# #             x=[]
# #             y=[]
# # 
# #             for i in range(len(all_cent)):
# #                 if np.absolute(all_cent[i] - np.median(all_cent)) <level:
# #                     x.append(i)
# #                     y.append(all_cent[i])
# #         
# #             all_cent = np.column_stack((x,y))
# #         
# #         else:
# #             all_cent = np.column_stack((np.linspace(0, len(all_cent), len(all_cent), dtype = 'uint16'), all_cent))
# # =============================================================================
#         
#         #plt.plot(all_cent[:,0], all_cent[:,1], 'bo')
#                     
#         return all_cent
# =============================================================================

        
        
        
        
        
        
    
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-




# ========Convention: note the difference from PEP8 for variables!=============
# Naming:
#   * classes MixedUpperCase
#   * varables lowerUpper _or_ lower
#   * functions and methods underscore_separated _or_ lower
# =============================================================================

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from maximus48.SSIM import *
from maximus48.SSIM_sf2 import divide
from maximus48.SSIM_sf import *

from maximus48.var import *
from maximus48.polar import *
from maximus48.polar import polar_image

from maximus48.monochromaticCTF import single_distance_CTF
from maximus48.multiCTF2 import *
from maximus48.sidi_phare import Paganin, MBA, BAC, anka_filter



__module__ = "maximus48"
__author__ = \
    "Maxim Polikarpov (European Molecular Laboratory Lab"
__email__ = "polikarpov.maxim@mail.ru"
__version__ = '0.3.4'
__date__ = "1 Aug 2019"
__license__ = "MIT license"

__all__ = ['im_folder',
           'read_stack',
           'show',
           'imrescale',
           'imrescale_interactive',
           'make_video',
           'wavelen',
           
           'SSIM',
           'single_distance_CTF',
           'multi_distance_CTF',
           'shift_image',
           'shift_image_CTF'
           ]
           











#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug  6 17:46:49 2018

@author: mpolikarpov
"""


import matplotlib.pyplot as plt
import cv2
import os
#from PIL import Image
import numpy as np
from numpy import mean, square, sqrt
from numpy.fft import fft2, fftshift
from joblib import Parallel, delayed



def im_folder(path):
    """
    lists images in the folder 
    
    Parameters
    __________
    path : str
    """
    
    fileformat = 'ppm','PPM','tiff','TIFF','tif','TIF','png','PNG', 'raw', 'jpg', 'JPG'
    curfol = os.getcwd()
    
    os.chdir(path)
    imfiles = os.listdir(path)
    imlist = [filename for filename in imfiles if filename.endswith(fileformat)]
    imlist.sort()
    os.chdir(curfol)
    
    return(imlist)
    
    
    
    
    
def images_folder(path, dtype, ROI = None, data_folder_name = 'data', flatfield_folder_name = 'flatfield', asarray = True):
    
    """
    reads all images in the subfolders of the folder 'path': 

        
    Parameters
    __________
    path : str
        path to the main directory
    dtype: str
        data-type
    ROI  : tuple 
        (typically 4 numbers with (x,y,x2,y2), optional
    data_folder_name : str
        name of the folder with data, optional
    flatfield_folder_name: str
        name of the folder with flatfield, optional      
    asarray : boolean
        if True, will return an ndarray.
        Attention! to return a 3D array correctly, you need to be sure that each folder has equal number of files  
    """
    
    data=[]
    flatfield=[]
    #data_folders=[]
    #flatfield_folders=[]
    
    for root, dirs, files in os.walk(path):        
        [data.append(read_stack2(root+'/'+name, dtype, ROI, asarray = True)) for name in dirs if data_folder_name in name]
        [flatfield.append(read_stack2(root+'/'+name, dtype, ROI, asarray = True)) for name in dirs if flatfield_folder_name in name]
        #[data_folders.append(root+'/'+name) for name in dirs if data_folder_name in name]
        #[flatfield_folders.append(root+'/'+name) for name in dirs if flatfield_folder_name in name]
        
    if asarray:
        data = np.asarray(data)
        flatfield = np.asarray(flatfield)
    
    return data, flatfield
    


def images_folder2(path, data_folder_name = 'data', flatfield_folder_name = 'flatfield'):
    
    """
    finds all images in the subfolders of the folder 'path': 

        
    Parameters
    __________
    path : str
        path to the main directory
    data_folder_name : str
        name of the folder with data, optional
    flatfield_folder_name: str
        name of the folder with flatfield, optional      
    """
    
    data_folders=[]
    flatfield_folders=[]
    
    for root, dirs, files in os.walk(path):        
        [data_folders.append(root+'/'+name) for name in dirs if data_folder_name in name]
        [flatfield_folders.append(root+'/'+name) for name in dirs if flatfield_folder_name in name]
        
    
    return data_folders, flatfield_folders







def read_image(path, dtype='float32', ROI = None, opencv = False):
    
    """
    reads one image as 2D-array
    
    Parameters
    __________
    path : str
        full path to the file
    dtype: str
        data-type
    ROI: tuple
        region of interest to read, (typically 4 numbers with (x,y,x2,y2), optional
    if cv2: True
        it will use cv2 module to read files
    """
    if opencv:
        image = cv2.imread(path,cv2.IMREAD_UNCHANGED).astype(dtype)
    else:
        image = plt.imread(path).astype(dtype)
    if ROI: 
        image = image[ROI[1]:ROI[3], ROI[0]:ROI[2]]
    
    return image
    
   
    
    
def read_stack(path, dtype, ROI=None):
    """
    reads all images in the folder  with DXchange module
        advantages: fast
        disadvantages: file names should be the same, indexing starts from 0 (first file)
        
    Parameters
    __________
    path : str
        path to the file
    dtype: str
        data-type
    ROI  : tuple 
        (typically 4 numbers with (x,y,x2,y2), optional
    """
    
    import dxchange 
    
    imlist = im_folder(path)
    os.chdir(path)
    
    n = len(imlist)  
    if ROI:
        proj = dxchange.reader.read_tiff_stack(imlist[0], range(n), digit = len(str(n)), slc = ((ROI[1],ROI[3],1),(ROI[0],ROI[2],1))).astype(dtype)
    else:
        proj = dxchange.reader.read_tiff_stack(imlist[0], range(n), digit = len(str(n))).astype(dtype)
    return(proj)


    
def read_stack2(path, dtype, ROI=None, 
                asarray=True, 
                opencv = False, 
                multiprocess = False, ncore = 5):
    """
    reads all images in the folder  with pillow module: 
        advantages: files can have any names; clear logic
        disadvantages: can be slower than read_stack
        
    Parameters
    __________
    path : str
        path to the file
    dtype: str
        data-type
    ROI  : tuple 
        (typically 4 numbers with (x,y,x2,y2), optional
    asarray : boolean
        if true, returns as array, otherwise- as a list
    if cv2: True
        it will use cv2 module to read files
    """
    
    
    fileformat = 'ppm','PPM','tiff','TIFF','tif','TIF','png','PNG', 'raw'      # formats available
    imlist = [filename for filename in os.listdir(path) if filename.endswith(fileformat)]
    imlist.sort()
    
    if multiprocess:
        out = (Parallel(n_jobs=ncore, prefer='threads')(delayed(read_image)(path+i, dtype, ROI, opencv) for i in imlist))
    
    else:
        out=[]
        for i in imlist:
            out.append(read_image(path+i, dtype, ROI, opencv))
            print('reading file: ', i)
    
    
    if asarray:
        out = np.asarray(out)
        
    return out
    
    

def show(image):
    
    """
    shows the 2D image 
    
    Parameters
    __________
    image : 2D array
    """
    
    plt.figure(num=None, figsize=(20, 20), facecolor='w', edgecolor='k')
    plt.imshow(image, cmap='gray')
    return;
    
    
def fourimage(image):
    """
    does inverse fourier transform and shows the image
    
    Parameters
    __________
    data : ndarray 
        input image data 2D or 3D array
    bitnum: int 
        number of bits to save. by default is 8 bit
    """
    ft = fft2(image)
    out = np.log(abs(fftshift(ft)))
    show(out)
    return 
    
    
def imrescale(data, bitnum=16):
    """
    increases brightness/contrast and returns it as int array
    
    Parameters
    __________
    data : ndarray 
        input image data 2D or 3D array
    bitnum: int 
        number of bits to save. by default is 8 bit
    """
    bit = 2**bitnum-1
    out = (data-np.min(data))*bit/(np.max(data)-np.min(data))
    out = out.astype('uint'+str(bitnum))
    return out
    

def imrescale_interactive(data, bitnum=8):
    """
    increases brightness/contrast by user-defined value and returns it as int array
    
    Parameters
    __________
    data : ndarray 
        input image data 3D array
    bitnum: int 
        number of bits to save. by default is 8 bit
    """
    
    Flag = True
    bit = 2**bitnum-1
    
    while Flag:
        #Amp is the input from user
        Amp = float(input("Please enter amplification factor: "))
        delta = Amp * sqrt(mean(square(data))) 
        test = (data - mean(data)+delta)*bit//(2*delta)
        np.clip(test, 0, bit, test)
        show(test)
        plt.show()
        
        '''
        plt.figure(num=None, figsize=(12, 14), facecolor='w', edgecolor='k')
        plt.imshow(test, cmap='Greys_r')
        plt.show()
        '''
        #check if the user is satisfied
        flag2 = int(input("Type '1' if ok or '0' to redo the normalization:  "))
        if flag2 == 1:
            data = (data - mean(data)+delta)*bit//(2*delta)                        # dynamical range for the substraction - type 1.65 if you want to leave central 90% of signal range, type 2 for 90% and 2.6 for 99%
            np.clip(data, 0, bit, data)                                        # makes all the pixels that are bigger that 255 equal 255 and that are less 0 to be 0 (example is for 8 bit)
            data = np.array(np.round(data),dtype = 'uint'+str(bitnum))                           # Round values in array and cast as whatever-bit integer
        
            Flag = False
    return data
        
    
def make_video(stack, video_name='video_stack.avi', path=os.getcwd()):

    """
    Makes a JPG video out of the image stack
    
    Parameters
    __________
    stack : ndarray
        stack of images where the 0 axis corresponds to the image index
    video_name : str, optional
        the name of the output file
    path : str, optional
        path where the file should be saved
    """
    
    video = cv2.VideoWriter(path+'/'+video_name, cv2.VideoWriter_fourcc('M','J','P','G'), 15, (stack.shape[1], stack.shape[2]), isColor = True)
    for i in range(stack.shape[0]):
        file = stack[i,:,:]
        #video.write(file)
        video.write(cv2.merge([file,file,file]))
    video.release()


def wavelen(energy):
    """
    Calculates the wavelength out of Energy in keV
    
    Parameters
    __________
    energy : int
    """
    
    h = 4.135667662 * 1e-18 # plank constant, keV*sec
    c = 299792458           # speed of light , m/sec
    Y = (h*c)/energy
    return Y


def maximal_intensity(image, angle = None):
    
    """
    Метод максимальной интенсивности (томо)
    
    Parameters
    __________
    image : ndarray
        stack of images where the 0 axis corresponds to the tomo slice
    angle : int
        угол обзора 

    """
    from scipy.ndimage import rotate
    if angle:
        image = rotate(image, angle, axes=(2,1))
    IM_MAX= np.max(image, axis=2)

    return IM_MAX









